/*
    <ClientSample.c>>
ңУǺץ
     ꥨ ͥåȥңУǥ <ͥåȥط饤֥ꥵץ륽>
     Version 0.001, Dec 2002
     ------------------------------------------------------------
     Copyright (C) 2002ǯ naoki kuwamori<naoki@kuwamori.com>

     ܥץϥե꡼եȥǤʤϡFree Software Foundation
     ɽGNU ̸ͭѵΡ֥С󣲡װϤʹߤγƥС
     椫餤줫򤷡ΥС˽äܥץ
     ۤޤѹ뤳ȤǤޤ

    ʳΥ饤󥹤˾ξϡ᡼뤯.
    <naoki@kuwamori.com>


    Laurier Net work RPG server<game network library>
    Version 0.001, Dec 2002
    Copyright (C) 2002 Naoki Kuwamori <naoki@kuwamori.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or any
    later version.

    If you wish another license . 
    mail me <naoki@kuwamori.com> .

*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>

#include "libnet.h"
/*ϤäȤ
#include "laurel/libnet.h"
*/
int dummyfunc(char *command , char * readbuf , int read_len , int resfd);
int helthcheck(char *command , char * readbuf , int read_len , int resfd);
int chdlevel(char *command , char * readbuf , int read_len , int resfd);
int endreq(char *command , char * readbuf , int read_len , int resfd);
int OnTimeup(char *command , char * readbuf , int read_len , int resfd);
int shmctlreq(char *command , char * readbuf , int read_len , int resfd);
int rawdatarec(char *command , char * readbuf , int read_len , int resfd);
int mmaptest(char *command , char * readbuf , int read_len , int resfd);
int shmtest(char *command , char * readbuf , int read_len , int resfd);


#define DYMMYMES            "DYMMYMESGE"            /*ߡΥåݻ*/
#define HELTHCHK            "HELTHCHECK"            /*إ륹å*/
#define HELTHCHKOK          "HELTHCHKOK"            /*إ륹å*/
#define DBGLEVEL            "DEBUGLEVEL"            /*ǥХå٥륻å׵*/
#define ENDREQ              "ENDREQUEST"            /*λ׵*/
#define SHMCTLREQ           "SHMCNTLREQ"           /*ͭɣļ׵*/

#define NORMALSENDTEST		"NORMSETEST"			/*̾ƥ*/
#define LARGESENDTEST		"LARGESTEST"			/*̾ƥ*/
#define SMALLSENDTEST		"SMALLSTEST"			/*̾ƥ*/
#define MMAPTESTREQ "MMAPTEST__"
#define SHMEMTESTREQ    "SHMEMTEST_"


#define LAURELLOGSOCNAME    "/tmp/laurellog"
#define DEFAULTGAMENAME "laureltest"
#define CONFIGFILENAME "config_file.csv"
#define LAURELLOGFILENAME  "laurelsampcli.log"
#define LIBNET_DBG_LOGFUNC        150
#define STDIN				0

    int UNIXserverFD;
    int TCPserverFD;
    int UDPserverFD;
	
	static  int ctlshmid; /*ͭѣɣĥåΰ*/
	LP_shmunit * ctlshmadder;
	char * shmadder;
	char *adder;
	int testflag=0;

/*TODO ץ
 * ʴǥեեƤ
 * ¤Τ*/
typedef struct{
	char hostname[1024];
	char portnom[1024];
	char UnixSocName[1024];
	char MMapFilePath[1024];
	char Uportnom[1024];
	int  shmID;
}config_file;
	config_file testval;

/*TODO ץ
 * եեΥƥץ졼Ȥ롣*/
static CCSVTemplate tmpl_templateConfig_file[] = { 
    CSVTRECORD( offsetof( config_file, hostname),CSV_string,1023), 
    CSVTRECORD( offsetof( config_file, portnom),CSV_string,1023), 
    CSVTRECORD( offsetof( config_file, UnixSocName),CSV_string,1023),
    CSVTRECORD( offsetof( config_file, MMapFilePath),CSV_string,1023),
	CSVTRECORD( offsetof( config_file, Uportnom),CSV_string,1023),
	CSVTRECORD( offsetof( config_file, shmID),CSV_int,0),
    EndOfCSVTemplate
};


/*Хåؿ򥻥åȤ*/
/*Хåؿ򥻥åȤ빽¤Τ*/
typedef struct {
	    char event[11];             /* number of event */
		int (*func)();  /* function address */
}EVTBL;
/*TODO ץ
 * ٥ȥޥåפ򤳤ˤ
 * ϥ٥ȤФ륳Хåؿ򤳤Ͽ롣*/
EVTBL eventkye[20] = {
         /* ٥             ؿ*/
	{DYMMYMES,               dummyfunc},      /*ߡؿ*/
	{HELTHCHK,               helthcheck},     /*إ륹å׵*/
	{HELTHCHKOK,             helthcheck},     /*إ륹å*/
	{DBGLEVEL,               chdlevel},       /*ǥХå٥ѹ׵*/
	{ENDREQ,                 endreq},         /*λ׵*/
	{LIBNET_CMD_POLLTIMEUP,			 OnTimeup},
	{SHMCTLREQ,              shmctlreq},         /*ͭꥳȥΰ*/
	{LIBNET_CMD_RAWDATARECEVE,	 rawdatarec},	/*rawμؤ*/
	{MMAPTESTREQ,           mmaptest},
	{SHMEMTESTREQ,          shmtest},
	{"",                     NULL}
};

char * gamename(char * lgamename){
    static char ggamename[MAXPATHLEN];
    static int flag=0;
    if (flag==0){
        bzero(ggamename,MAXPATHLEN);
        strncpy(ggamename,DEFAULTGAMENAME,MAXPATHLEN);
        flag=1;
    }
    if (lgamename!=NULL) strncpy(ggamename,lgamename,MAXPATHLEN);
    return ggamename;
}

char * logsockname(char * socname){
    static char gsockname[MAXPATHLEN];
    static int flag=0;
    if (flag==0){
        bzero(gsockname,MAXPATHLEN);
        strncpy(gsockname,LAURELLOGSOCNAME,MAXPATHLEN);
        flag=1;
    }
    if (socname!=NULL) strncpy(gsockname,socname,MAXPATHLEN);
    return gsockname;
}


/*ɤ߽ФƥåȤ*/
/*TODO ץ
 * ư˴ؤϤ
 * man getopt_long 򻲾Ȥ뤳*/
int set_opt(int argc, char *argv[]){
    int c;
    while (1)
    {
        int option_index = 0;
        static struct option long_options[] =
        {
           {"version", 0, 0, 'v'},
           {"help", 0, 0, 'h'},
           {"logfilename", 1, 0, 'f'},
           {"logsocname", 1, 0, 's'},
           {"debug", 1, 0, 'd'},
           {"gamename", 1, 0, 'g'},
           {0, 0, 0, 0}
        };
        c = getopt_long (argc, argv, "vhf:s:d:g:t",
                    long_options, &option_index);
        if (c == -1) break;
        switch (c)
        {
           case 't':
			   testflag=1;
             break;
           case 'v':
                printf ("LaurelRPG server\n");
                exit (1);
             break;
           case 'h':
                printf ("LaurelRPG server\n");
                exit (1);
             break;
           case 'f':
                LP_setlogfile(optarg);
             break;
           case 's':
                logsockname(optarg);
             break;
           case 'd':
                LP_dbglvl(atoi(optarg));
             break;
           case 'g':
                gamename(optarg);
             break;

           case '?':
             break;
           default:
             printf ("?? getopt returned character code 0%o ??\n", c);
        }
    }

    if (optind < argc)
    {
     printf ("nonoption ARGVelements: ");
     while (optind < argc)
     printf ("%s ", argv[optind++]);
     printf ("\n");
    }
	return (optind);
}


int main (int argc, char *argv[])
{
    char readbuf[1025];
    int read_len;
    char command[20];
    int resfd;
    int cnt;
    int ret;
	FILE *fp;

    /*ǥեͤ򥻥åȤ*/
    LP_dbglvl(101);
    LP_setlogfile(LAURELLOGFILENAME);

    /*ᤷѿ˼Ǽ*/
    set_opt(argc,argv);

    /*ե졼եɤ߹*/

    /*եΥץ*/
    if ((ret=  LP_log(LIBNET_DBG_EMERG ,"start loging\n" )) < 0){
        printf("logopnen error abort\n");
        exit (0);
    }

/*TODO ץ
 * CSVǤ줿ե졼եɤ߹߽*/
	fp = fopen (CONFIGFILENAME,"r");
	if (fp==NULL){
		printf("file open error\n");
		exit(0);
	}
	while ( fgets( readbuf, sizeof( readbuf ), fp ), !feof( fp ) ) {
		if (readbuf[0]=='#')continue; /*ȤλϹԤκǽʸˡ*/
		memset (&testval , 0 ,sizeof(config_file));
		if (FALSE==GetCSV( &testval, readbuf, tmpl_templateConfig_file )){
			printf("error of read cvs file\n");
		}   
		printf("print nom[%s] string[%s] Mmap[%s]\n", testval.portnom , testval.UnixSocName,testval.MMapFilePath);
		break;
	}


/*TODO ץ
 * ˥ååȡ̾դѥסˤΥ饤ȥץ*/
    /*ͥåȥΥץ(˥ååȤΥ饤ȥץ)*/
    UNIXserverFD = LP_usoc_client_open(testval.UnixSocName );
    if (UNIXserverFD <= 0){
        LP_log(LIBNET_DBG_EMERG,"LP_usoc_client_open error [%s] [%s] [%d]\n",testval.UnixSocName, strerror(errno) , ret);
        LP_endproc(1);
    }
	
/*TODO ץ
 * ԣãХåȤΥ饤ȥץ*/
    TCPserverFD = LP_tcp_client_open(testval.hostname,testval.portnom );
    if (TCPserverFD <= 0){
        LP_log(LIBNET_DBG_EMERG,"LP_tcp_client_open error [%s] [%s] [%d]\n",testval.portnom, strerror(errno) , ret);
        LP_endproc(1);
    }
/*TODO ץ
 * UDPåȤΥ饤ȥץ*/
    LP_log(LIBNET_DBG_EMERG,"LP_udp_client_open [%s] [%s]\n",testval.Uportnom, strerror(errno) );
    UDPserverFD = LP_udp_client_open(testval.hostname,testval.Uportnom );
    if (UDPserverFD <= 0){
        LP_log(LIBNET_DBG_EMERG,"LP_udp_client_open error [%s] [%s] [%d]\n",testval.Uportnom, strerror(errno) , ret);
        LP_endproc(1);
    }
    /*ͭγ*/
/*TODO ץ
 * ͣͣͭΥץ*/
    (char *)ret = LP_mopen(testval.MMapFilePath);
    if ((char *)ret == NULL){
        LP_log (LIBNET_DBG_EMERG,"cant open mmmap\n");
        LP_endproc(LP_NG);
    }
    adder = (char *)ret;

/*TODO ץ*/
    /*ФȤΥå³*/
    ret = LP_loginit( argv[0] , LAURELLOGSOCNAME);
    if (ret < 0){
        LP_log (LIBNET_DBG_EMERG,"cant connect log server\n");
    }


/*TODO ץ*/
    /*λΤǥǡˤʤ*/
	if (testflag != 1){
		ret = fork();
		if (ret > 0){
			if (ret==1){
				LP_log (LIBNET_DBG_EMERG,"cant fork child prosses\n");
			}
			exit (1);
		}
/* ȤɸϤңףƣϿư򸡾ڤ*/
	}else{
		LP_log (LIBNET_DBG_EMERG,"set Raw FD\n");

		LP_addRawPollFD(STDIN);

	}
	/*ͭɣĤ򥵡Ф˥䤤碌ʶͭȤɬܡ*/
	LP_send(TCPserverFD,SHMCTLREQ,"",0);

    /*ᥤ롼*/
    while (1){
        memset (readbuf , 0 , 1024);
        ret = LP_Poll (command ,  readbuf , 1024, &read_len , 10000 , &resfd );
        readbuf[1023]=0;
        LP_log(1050, "ret %d [%.10s] \n" , ret , command );
        for (cnt=0;cnt<20;cnt++){
            if (eventkye[cnt].func==NULL)break;
                /*ޥʸӤפ*/
            if (strncmp(command ,eventkye[cnt].event,COMMANDLEN)==0){
                /*ХåؿƤӽФ*/
                ret = eventkye[cnt].func(command ,readbuf,read_len , resfd);
            }
        }
    }
    return LP_OK;
}

/*ߡؿ*/
int dummyfunc(char *command , char * readbuf , int read_len , int resfd){
    LP_log(LIBNET_DBG_LOGFUNC , "dummyfunc\n");

    return LP_OK;
}
/*إ륹åؿ׵ξ*/
int helthcheck(char *command , char * readbuf , int read_len , int resfd){
    LP_log(LIBNET_DBG_LOGFUNC , "helthcheck\n");

    /*ֿ*/
    if (memcmp(HELTHCHK,command,COMMANDLEN)==0){
        LP_send(resfd,HELTHCHKOK,"",0);
    }else{
		LP_send(UNIXserverFD,DYMMYMES , "",0);
	}
    return LP_OK;
}
/*ǥХå٥ѹ׵*/
int chdlevel(char *command , char * readbuf , int read_len , int resfd){
    int rev;
    LP_log(LIBNET_DBG_LOGFUNC , "chdlevel\n");
    /*ӥåǥб*/
    rev = ntohl (*((int *)readbuf));
    LP_dbglvl(rev);
    return LP_OK;
}
/*λ׵*/
int endreq(char *command , char * readbuf , int read_len , int resfd){
    LP_log(LIBNET_DBG_LOGFUNC , "endreq\n");
    LP_fd_allclose();
    LP_log(-1,"end log file\n");
	exit(0);
}

int OnTimeup(char *command , char * readbuf , int read_len , int resfd){
    LP_log(LIBNET_DBG_LOGFUNC , "OnTimeup\n");

	/*إ륹å*/
	/*LP_send( TCPserverFD, HELTHCHK , "",0);*/

}

/*ͭꥳȥɣļؿ*/
int shmctlreq(char *command , char * readbuf , int read_len , int resfd){
	memcpy ((char *)&ctlshmid , readbuf , sizeof(int));
	LP_log(LIBNET_DBG_LOGFUNC , "shmctlreq %d\n" , ctlshmid);

	/*ȥɣĤǤ麣٤*/
	ctlshmadder = LP_shm_init (ctlshmid);
	shmadder =  (char *)LP_shm_attach( testval.shmID);
	LP_log(LIBNET_DBG_LOGFUNC , "shmctlreq:shmadder [%x] %d\n",shmadder,shmadder);

	return LP_OK;
}

/*ɸϤΥޥɼջ*/
int rawdatarec(char *command , char * readbuf , int read_len , int resfd){
	char buf[1024];
	char sendbuf[4000];
	int ret,cnt;
	int sum;
	LP_log(LIBNET_DBG_LOGFUNC , "rawdatarec\n");
	/*****************************************************/
	/*			Ƽƥȥɤ					*/
	/*****************************************************/

	ret = read (STDIN , buf,1024);
	if (ret == 0){
		LP_log(LIBNET_DBG_LOGFUNC , "rawdatarec:read length 0\n");
		return LP_OK;
	}


	/*إ륹å*/
	if (NULL!= strstr(buf,"HELTHCHK")){ 
		LP_send( TCPserverFD, HELTHCHK , "",0);
		
	}else if(NULL!= strstr(buf,"SENDTEST")){
	/*ƥ*/
		if(NULL!= strstr(buf,"UDP")){
			LP_log(LIBNET_DBG_LOGFUNC , "rawdatarec:UDP send test start\n");
			/*UDP*/
			LP_log(LIBNET_DBG_LOGFUNC , "rawdatarec:UDP normal send \n");
			sprintf(sendbuf,"rawdatarec:UDP send test start");
			LP_send(UDPserverFD,NORMALSENDTEST ,sendbuf ,strlen(sendbuf));
			/*礭ʥǡʣХȰʾ*/
			LP_log(LIBNET_DBG_LOGFUNC , "rawdatarec:UDP big data send \n");
			for(sum=0,cnt=0;cnt<3500;cnt++){
				int i;
				i=(int) (10.0*rand()/(RAND_MAX+1.0))+060;
				sendbuf[cnt]=i;
				i=i%256;
/*				printf ("%c",i);*/
				sum=sum+i;
			}
			snprintf(&(sendbuf[3500]),500,"%d",sum);
			LP_send(UDPserverFD,LARGESENDTEST,sendbuf ,4000);
			/*ʥǡʣʥѥå礬Ԥ١*/
			LP_log(LIBNET_DBG_LOGFUNC , "rawdatarec:UDP small data lot send \n");
			for(cnt=0;cnt<500;cnt++){
				sprintf(sendbuf,"%d",cnt);
				LP_send(UDPserverFD,SMALLSENDTEST,sendbuf ,strlen(sendbuf));
			}


		}else if(NULL!= strstr(buf,"UNIX")){
			LP_log(LIBNET_DBG_LOGFUNC , "rawdatarec:UNIX send test start\n");

			/*գΣɣإå*/
			sprintf(sendbuf,"rawdatarec:UNIX send test start");
			LP_send(UNIXserverFD,NORMALSENDTEST ,sendbuf ,strlen(sendbuf));
			LP_log(LIBNET_DBG_LOGFUNC , "rawdatarec:UNIX normal send \n");

			/*礭ʥǡʣХȰʾ*/
			LP_log(LIBNET_DBG_LOGFUNC , "rawdatarec:UNIX big data send \n");
			for(sum=0,cnt=0;cnt<3500;cnt++){
				int i;
				i=(int) (10.0*rand()/(RAND_MAX+1.0))+060;
				sendbuf[cnt]=i;
				i=i%256;
/*				printf ("%c",i);*/
				sum=sum+i;
			}
			snprintf(&(sendbuf[3500]),500,"%d",sum);
			LP_send(UNIXserverFD,LARGESENDTEST,sendbuf ,4000);

			/*ʥǡʣʥѥå礬Ԥ١*/
			LP_log(LIBNET_DBG_LOGFUNC , "rawdatarec:UNIX small data lot send \n");
			for(cnt=0;cnt<500;cnt++){
				sprintf(sendbuf,"%d",cnt);
				LP_send(UNIXserverFD,SMALLSENDTEST,sendbuf ,strlen(sendbuf));
			}
		}else{
			LP_log(LIBNET_DBG_LOGFUNC , "rawdatarec:TCP send test start\n");
			/*TCP*/
			LP_log(LIBNET_DBG_LOGFUNC , "rawdatarec:TCP normal send \n");
			sprintf(sendbuf,"rawdatarec:TCP send test start");
			LP_send(TCPserverFD,NORMALSENDTEST ,sendbuf ,strlen(sendbuf));
			/*礭ʥǡʣХȰʾ*/
			LP_log(LIBNET_DBG_LOGFUNC , "rawdatarec:TCP big data send \n");
			for(sum=0,cnt=0;cnt<3500;cnt++){
				int i;
				i=(int) (10.0*rand()/(RAND_MAX+1.0))+060;
				i=i%256;
				sendbuf[cnt]=i;
/*				printf ("%d",(i-060));*/
				sum=sum+i;
			}
			snprintf(&(sendbuf[3500]),500,"%d",sum);
			LP_send(TCPserverFD,LARGESENDTEST,sendbuf ,4000);
			/*ʥǡʣʥѥå礬Ԥ١*/
			LP_log(LIBNET_DBG_LOGFUNC , "rawdatarec:TCP small data lot send \n");
			for(cnt=0;cnt<500;cnt++){
				sprintf(sendbuf,"%d",cnt);
				LP_send(TCPserverFD,SMALLSENDTEST,sendbuf ,strlen(sendbuf));
			}
		}
		/**/
		
	}else if(NULL!= strstr(buf,"MMAPTESTSEND")){
		LP_log(LIBNET_DBG_LOGFUNC , "rawdatarec: mmap test\n");
		/**/
		for (cnt =0;cnt<2048;cnt++){
			(unsigned char)(adder[cnt])=cnt%256;
		}
		LP_send(TCPserverFD,MMAPTESTREQ,"+",1);
		sleep (1);
		for (cnt =0;cnt<2048;cnt++){
			(unsigned char)(adder[cnt])=(2048-cnt)%256;
		}
		LP_send(TCPserverFD,MMAPTESTREQ,"-",1);
		
	}else if(NULL!= strstr(buf,"MEMTESTSEND")){
	/*ͭϤƥ*/
		LP_log(LIBNET_DBG_LOGFUNC , "rawdatarec: shered memory test\n");

		/*ͭ*/
		for (cnt =0;cnt< 0x1000000;cnt++){
			(unsigned char)(shmadder[cnt])=cnt%256;
		}
		LP_send(TCPserverFD,SHMEMTESTREQ,"+",1);
		sleep (1);
		for (cnt =0;cnt< 0x1000000;cnt++){
			(unsigned char)(shmadder[cnt])=(0x1000000-cnt)%256;
		}
		LP_send(TCPserverFD,SHMEMTESTREQ,"-",1);
		/**/
		
	}else if(NULL!= strstr(buf,"MEMTESTSIZE")){
	/*ͭꥵƥ*/
		LP_log(LIBNET_DBG_LOGFUNC , "rawdatarec: shered memory test\n");
		/*꤮ޤǽ*/
		for (cnt =0;cnt< 0x1000000;cnt++){
			(unsigned char)(shmadder[cnt])=cnt%256;
		}
		LP_send(TCPserverFD,SHMEMTESTREQ,"+",1);
		sleep (1);
		for (cnt =0;cnt< 0x1000000;cnt++){
			(unsigned char)(shmadder[cnt])=(0x1000000-cnt)%256;
		}
		LP_send(TCPserverFD,SHMEMTESTREQ,"-",1);
		/**/

	}else if(NULL!= strstr(buf,"HELP")){
	/**/
		printf ("HELTHCHK:\n");
		printf ("	to send helthcheck\n");
		printf ("\n");
		printf ("SENDTEST:TCP\n");
		printf ("	TCP send test\n");
		printf ("\n");
		printf ("SENDTEST:UDP\n");
		printf ("	UDP send test\n");
		printf ("\n");
		printf ("SENDTEST:UNIX\n");
		printf ("	UNIX send test\n");
		printf ("\n");
		printf ("MMAPTESTSEND:\n");
		printf ("\n");
		printf ("MEMTESTSIZE:\n");
		printf ("\n");
		printf ("\n");
	/**/
	/**/
	/**/
	/**/
	/**/
	}

	return LP_OK;
}

int mmaptest(char *command , char * readbuf , int read_len , int resfd){
    int cnt;
    LP_log(100 , "mmaptest: mmapreadtest %s\n",readbuf);

    if (memcmp(readbuf,"+",1)==0){
        for (cnt =0;cnt<2048;cnt++){
            if ((unsigned char)(adder[cnt])!=cnt%256){
                LP_log(70 , "mmaptest: error mem mispmatch + [%d]-[%d] [%d]\n",adder[cnt],cnt%256,cnt);
            }
        }
    }else{
        for (cnt =0;cnt<2048;cnt++){
            if ((unsigned char)(adder[cnt])!=(2048-cnt)%256){
                LP_log(70 , "mmaptest: error mem mispmatch - [%d]-[%d] [%d]\n",adder[cnt],(2048-cnt)%256,cnt);
            }
        }
        for (cnt =0;cnt<2048;cnt++){
            (unsigned char)(adder[cnt])=cnt%256;
        }
        LP_send(resfd,MMAPTESTREQ,"+",1);
    }

	return LP_OK;
}
int shmtest(char *command , char * readbuf , int read_len , int resfd){
    int cnt;
    LP_log(100 , "shmtest: shmreadtest %s\n",readbuf);

    if (readbuf[0]=='+'){
        for (cnt =0;cnt<0x1000000;cnt++){
            if ((unsigned char)(shmadder[cnt])!=cnt%256){
                LP_log(70 , "shmtest: error mem mispmatch + [%d]-[%d] [%d]\n",shmadder[cnt],cnt%256,cnt);
            }
        }
    }else{
        for (cnt =0;cnt<0x1000000;cnt++){
            if ((unsigned char)(shmadder[cnt])!=(0x1000000-cnt)%256){
                LP_log(70 , "shmtest: error mem mispmatch - [%d]-[%d] [%d]\n",shmadder[cnt],(2048-cnt)%256,cnt);
            }
        }
        for (cnt =0;cnt< 0x1000000;cnt++){
            (unsigned char)(shmadder[cnt])=cnt%256;
        }
        LP_send(resfd,SHMEMTESTREQ,"+",1);
    }
	return LP_OK;
}

